import os
from datetime import datetime, timedelta, timezone
from urllib.parse import urlparse

from biolib._shared.types import ResourceDetailedDict
from biolib.api import client as api_client
from biolib.biolib_binary_format.utils import RemoteEndpoint
from biolib.biolib_logging import logger
from biolib.typing_utils import Optional


class DataRecordRemoteStorageEndpoint(RemoteEndpoint):
    def __init__(self, uri: str):
        self._uri: str = uri
        self._expires_at: Optional[datetime] = None
        self._presigned_url: Optional[str] = None

    def get_remote_url(self) -> str:
        if not self._presigned_url or not self._expires_at or datetime.now(timezone.utc) > self._expires_at:
            resource_response: ResourceDetailedDict = api_client.get(
                path='/resource/',
                params={'uri': self._uri},
            ).json()

            version = resource_response.get('version')
            assets = version.get('assets') if version else None
            if not assets:
                raise Exception(f'Resource "{self._uri}" has no downloadable assets')

            download_url = assets['download_url']
            app_caller_proxy_job_storage_base_url = os.getenv('BIOLIB_CLOUD_JOB_STORAGE_BASE_URL', '')
            if app_caller_proxy_job_storage_base_url:
                parsed_url = urlparse(download_url)
                self._presigned_url = f'{app_caller_proxy_job_storage_base_url}{parsed_url.path}?{parsed_url.query}'
            else:
                self._presigned_url = download_url

            self._expires_at = datetime.now(timezone.utc) + timedelta(minutes=8)
            logger.debug(
                f'DataRecord "{self._uri}" fetched presigned URL ' f'with expiry at {self._expires_at.isoformat()}'
            )

        return self._presigned_url



from biolib.biolib_logging import logger_no_user_data
from biolib.typing_utils import Optional, TypedDict
from biolib.utils.cache_state import CacheState

# TODO: Save job keys in the user state instead of a separate state file
# UuidStr = str
# class JobStateType(TypedDict):
#     job_uuid: UuidStr
#     aes_key: Optional[str]


class UserStateType(TypedDict):
    refresh_token: Optional[str]
    # jobs: Dict[UuidStr, JobStateType]


class UserState(CacheState[UserStateType]):
    def __init__(self) -> None:
        super().__init__(fail_fast_on_lock_acquire=True)
        self._is_in_memory_only: bool = False

    @property
    def _state_path(self) -> str:
        return f'{self._user_cache_dir}/user-state.json'

    def _get_default_state(self) -> UserStateType:
        return UserStateType(refresh_token=None)

    def __enter__(self) -> UserStateType:
        if self._is_in_memory_only:
            if self._state is None:
                self._state = self._get_default_state()
            return self._state
        try:
            return super().__enter__()
        except Exception as error:
            logger_no_user_data.warning(
                f'UserState: Could not access state file, continuing with in-memory state only. '
                f'Login state will not persist across Python processes. Error: {error}'
            )
            self._is_in_memory_only = True
            if self._state is None:
                self._state = self._get_default_state()
            return self._state

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if self._is_in_memory_only:
            return
        try:
            super().__exit__(exc_type, exc_val, exc_tb)
        except Exception as error:
            logger_no_user_data.warning(
                f'UserState: Could not write state file. '
                f'Login state will not persist across Python processes. Error: {error}'
            )
            self._is_in_memory_only = True



import json
from typing import Any, Dict

from biolib import api
from biolib._shared.types import ResourceDetailedDict
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_api_client.biolib_app_api import _get_resource_uri_from_str
from biolib.biolib_logging import logger


class Index:
    def __init__(self, _internal_state: ResourceDetailedDict):
        self._state = _internal_state

    def __repr__(self) -> str:
        return f'Index: {self._state["uri"]}'

    @property
    def uri(self) -> str:
        return self._state['uri']

    @property
    def id(self) -> str:
        return f'{self._state["account_uuid"]}.{self._state["uuid"]}'.replace('-', '_')

    @staticmethod
    def get_by_uri(uri: str) -> 'Index':
        normalized_uri = _get_resource_uri_from_str(uri)
        response: ResourceDetailedDict = api.client.get(path='/resource/', params={'uri': normalized_uri}).json()
        if response['type'] != 'index':
            raise Exception(f'Resource "{response["uri"]}" is not an Index')
        return Index(_internal_state=response)

    @staticmethod
    def create(uri: str, config: Dict[str, Any]) -> str:
        BiolibApiClient.assert_is_signed_in(authenticated_action_description='create an Index')

        response = api.client.post(
            path='/resources/indexes/',
            data={
                'uri': uri,
                'index_config': config,
            },
        )
        result = response.json()
        created_uri: str = result['uri']
        logger.info(f"Successfully created Index '{created_uri}'")
        return created_uri

    @staticmethod
    def create_from_config_file(uri: str, config_path: str) -> str:
        with open(config_path) as config_file:
            index_config = json.load(config_file)

        return Index.create(uri=uri, config=index_config)
